#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>
#include "types.h"
#include "bmp.h"

int load_image(const char *imgName, Image *img){
    BMP_FILE_HEADER fileHeader;
    BMP_INFO_HEADER infoHeader;
    BYTE *pData;
    BYTE dummy[3];
    int imgFd, i;
    int wStep, padding;

    imgFd = open(imgName, O_RDONLY);
    if(imgFd == -1) return -1;
    
    read(imgFd, &fileHeader.bfType, sizeof(fileHeader.bfType));
    if(fileHeader.bfType ^ 0x4D42) return -1;
    read(imgFd, &fileHeader.bfSize, sizeof(BMP_FILE_HEADER) - 2 * sizeof(fileHeader.bfType));
    read(imgFd, &infoHeader, sizeof(BMP_INFO_HEADER));
    
    img->bpp = infoHeader.biBitCount;
    img->height = infoHeader.biHeight;
    img->width = infoHeader.biWidth;
    wStep = img->width * (img->bpp >> 3);

    pData = img->pData = (BYTE *)malloc(img->height * wStep);
    padding = ((wStep + 3) & ~3) - wStep;
    lseek(imgFd, fileHeader.bfOffBits, SEEK_SET);
    
    for(i = img->height - 1;i ^ -1;--i){
        read(imgFd, pData + i * wStep, wStep);
        read(imgFd, dummy, padding);
    }

    close(imgFd);
    return 0;
}

int save_image(const char *imgName, const Image *img){
    BMP_FILE_HEADER fileHeader;
    BMP_INFO_HEADER infoHeader;
    BMP_RGB_QUAD rgb_q;
    BYTE *pData;
    BYTE dummy[3];
    int imgFd, i;
    int wStep, padding;

    fileHeader.bfType = 0x4D42;
    fileHeader.bfSize = img->width * img->height * (img->bpp >> 3) + (img->bpp == 8) * sizeof(BMP_RGB_QUAD) * 256;
    fileHeader.bfOffBits = sizeof(BMP_FILE_HEADER) + sizeof(BMP_INFO_HEADER) + (img->bpp == 8) * sizeof(BMP_RGB_QUAD) * 256;
    
    infoHeader.biSize = sizeof(BMP_INFO_HEADER);
    infoHeader.biWidth = img->width;
    infoHeader.biHeight = img->height;
    infoHeader.biPlanes = 1;
    infoHeader.biBitCount = img->bpp;
    infoHeader.biCompression = 0;
    infoHeader.biSizeImage = img->width * img->height;
    infoHeader.biXPelsPerMeter = 0;
    infoHeader.biYPelsPerMeter = 0;
    infoHeader.biClrUsed = 0;
    infoHeader.biClrImportant = 0;

    imgFd = open(imgName, O_WRONLY | O_CREAT | O_TRUNC, 0644);

    if(imgFd == -1) return -1;

    write(imgFd, &(fileHeader.bfType), sizeof(fileHeader.bfType));
    write(imgFd, &(fileHeader.bfSize), sizeof(BMP_FILE_HEADER) - 2 * sizeof(fileHeader.bfType));
    write(imgFd, &(infoHeader), sizeof(BMP_INFO_HEADER));
    
    if(infoHeader.biBitCount == 8){
        for(i=0;i<256;++i){
            rgb_q.rgbBlue = i;
            rgb_q.rgbGreen = i;
            rgb_q.rgbRed = i;
            rgb_q.rgbReserved = 0;
            write(imgFd, &rgb_q, sizeof(BMP_RGB_QUAD));
        }
    }
    
    wStep = img->width * (img->bpp >> 3);
    padding = ((wStep + 3) & ~3) - wStep;
    pData = img->pData;
    
    if(!pData) return -1;
    lseek(imgFd, fileHeader.bfOffBits, SEEK_SET);

    for(i = img->height - 1;i ^ -1;--i){
        write(imgFd, pData + i * wStep, wStep);
        write(imgFd, dummy, padding);
    }

    close(imgFd);
    return 0;
}

int create_image(Image *img, LONG width, LONG height, WORD bpp, BYTE val){
    int r, c, wStep;

    wStep = width * (bpp >> 3);
    img->width = width;
    img->height = height;
    img->bpp = bpp;
    img->pData = (BYTE *)malloc(height * wStep);
    if(val >= 0 && val <= 255){
        for(r = 0;r < height;++r){
            BYTE *pData = img->pData + r * wStep;
            for(c = 0;c < wStep;++c){
                pData[c] = 0;
            }
        }
    }
    return 0;
}

int copy_image(const Image *img1, Image *img2){
    int h, wStep;
    int r, c;
    
    create_image(img2, img1->width, img1->height, img1->bpp, -1);
    
    wStep = img1->width * (img1->bpp >> 3);
    h = img1->height;
    for(r = 0;r < h;++r){
        BYTE *pData1 = img1->pData + r * wStep;
        BYTE *pData2 = img2->pData + r * wStep;
        for(c = 0;c < wStep;++c){
            pData2[c] = pData1[c];
        }
    }
    return 0;
}

void rgb565_to_rgb888(const Image *rgb565, Image *rgb888){
    int i, j, k;
    int w, h, wStep;
    BYTE *pData1;
    WORD *pData2;
    
    w = rgb565->width;
    h = rgb565->height;
    wStep = w * 3;

    for(i=0;i<240;++i){
        pData1 = rgb888->pData + i * wStep;
        pData2 = (WORD *)(rgb565->pData + i * (w * 2));
        for(j=0, k=0;j<wStep;j += 3, ++k){
            pData1[j + 0] = (BYTE)((pData2[k] & 0x001F) << 3);
            pData1[j + 1] = (BYTE)((pData2[k] & 0x07E0) >> 3);
            pData1[j + 2] = (BYTE)((pData2[k] & 0xF800) >> 8);
        }
    }
}

void rgb2gray(Image *img1, Image *img2){
    int c, r, pos;
    int h = img1->height;
    int w = img1->width;
    int wStep = w * (img1->bpp >> 3);
    BYTE *pData1;
    BYTE *pData2;
    
    create_image(img2, w, h, 8, -1);
    for (r = 0; r < h; ++r){
        pos = 0;
        pData1 = img1->pData + r * wStep;
        pData2 = img2->pData + r * w;
        for (c = 0; c < w; ++c, pos += 3){
            pData2[c] = 0.114*pData1[pos + 0] + 0.587*pData1[pos + 1] + 0.299* pData1[pos + 2];
        }
    }
}

void yuv422_to_rgb888(int width, int height, const BYTE* yuv420, BYTE* pRGB24)
{
    int nSize = height * (width / 2);
    int idxSrc = 0;
    int idxDst = 0;
    int i;
    BYTE y1, u, y2, v;
    BYTE r, g, b;

    for (i=0 ; i<nSize ; i++)
    {
        y1 = yuv420[idxSrc  ];
        u  = yuv420[idxSrc+1];
        y2 = yuv420[idxSrc+2];
        v  = yuv420[idxSrc+3];

        // 첫 번째 픽셀
        b = CLIP((76284 * (y1 - 16) + 132252 * (u - 128)) >> 16);
        g = CLIP((76284 * (y1 - 16) -  53281 * (v - 128)-  25625 * (u - 128)) >> 16);
        r = CLIP((76284 * (y1 - 16) + 104595 * (v - 128)) >> 16);

        pRGB24[idxDst  ] = b;
        pRGB24[idxDst+1] = g;
        pRGB24[idxDst+2] = r;

        idxDst += 3;

        // 두 번째 픽셀
        b = CLIP((76284 * (y2 - 16) + 132252 * (u - 128)) >> 16);
        g = CLIP((76284 * (y2 - 16) -  53281 * (v - 128) -  25625 * (u - 128)) >> 16);
        r = CLIP((76284 * (y2 - 16) + 104595 * (v - 128)) >> 16);

        pRGB24[idxDst  ] = b;
        pRGB24[idxDst+1] = g;
        pRGB24[idxDst+2] = r;

        idxDst += 3;
        idxSrc += 4;
    }
}

int free_image(const Image *img){
    free(img->pData);
    return 0;
}
