#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/videodev2.h>
#include "video_capture.h"

int xioctl(int fd, int request, void *arg){
    int r;
 
    do{
        r = ioctl(fd, request, arg);
    }while (-1 == r && EINTR == errno);
 
    return r;
}

int stream_on(VideoCapture *cap){
    struct v4l2_buffer buf = {0};

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = 0;
    if(-1 == xioctl(cap->fd, VIDIOC_STREAMON, &buf.type)){
        perror("Start Capture");
        return 1;
    }
    return 0;
}

int stream_off(VideoCapture *cap){
    struct v4l2_buffer buf = {0};

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(-1 == xioctl(cap->fd, VIDIOC_STREAMOFF, &buf.type)){
        perror("Start Capture");
        return 1;
    }
    return 0;
}

int init_mmap(VideoCapture *cap){
    struct v4l2_requestbuffers req = {0};
    struct v4l2_buffer buf = {0};

    req.count = 1;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_MMAP;
    if (-1 == xioctl(cap->fd, VIDIOC_REQBUFS, &req)){
        perror("Requesting Buffer");
        return 1;
    }

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = 0;
    if(-1 == xioctl(cap->fd, VIDIOC_QUERYBUF, &buf)){
        perror("Querying Buffer");
        return 1;
    }

    cap->data = (unsigned char *)mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, cap->fd, buf.m.offset);
    cap->width = 640;
    cap->height = 480;
    cap->bpp = 16;
    printf("Length: %d\nAddress: %p\n", buf.length, cap->data);
    printf("camera image size: %d\n", buf.bytesused);
 
    return 0;
}

VideoCapture open_camera(const char *dev_name){
    VideoCapture cap;
    int fd;
    fd = open(dev_name, O_RDWR);
    if(fd == -1){
        cap.fd = fd;
        return cap;
    }
    cap.fd = fd;

    if(print_caps(&cap)){
        close(cap.fd);
        return cap;
    }

    if(init_mmap(&cap)){
        close(cap.fd);
        return cap;
    }

    if(stream_on(&cap)){
        close(cap.fd);
        return cap;
    }
    return cap;
}


int print_caps(VideoCapture *cap){
    struct v4l2_capability caps = {0};
    struct v4l2_cropcap cropcap = {0};
    struct v4l2_fmtdesc fmtdesc = {0};
    struct v4l2_format fmt = {0};
    char fourcc[5] = {0};
    char c, e;
    int support_grbg10 = 0;
 
    if (-1 == xioctl(cap->fd, VIDIOC_QUERYCAP, &caps)){
            perror("Querying Capabilities");
            return 1;
    }
 
    printf( "Driver Caps:\n"
            "  Driver: \"%s\"\n"
            "  Card: \"%s\"\n"
            "  Bus: \"%s\"\n"
            "  Version: %d.%d\n"
            "  Capabilities: %08x\n",
            caps.driver,
            caps.card,
            caps.bus_info,
            (caps.version>>16)&&0xff,
            (caps.version>>24)&&0xff,
            caps.capabilities);
 
 
    
    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == xioctl (cap->fd, VIDIOC_CROPCAP, &cropcap)){
            perror("Querying Cropping Capabilities");
            return 1;
    }
 
    printf( "Camera Cropping:\n"
            "  Bounds: %dx%d+%d+%d\n"
            "  Default: %dx%d+%d+%d\n"
            "  Aspect: %d/%d\n",
            cropcap.bounds.width, cropcap.bounds.height, cropcap.bounds.left, cropcap.bounds.top,
            cropcap.defrect.width, cropcap.defrect.height, cropcap.defrect.left, cropcap.defrect.top,
            cropcap.pixelaspect.numerator, cropcap.pixelaspect.denominator);
 
    
    fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    
    printf("  FMT : CE Desc\n--------------------\n");
    while (0 == xioctl(cap->fd, VIDIOC_ENUM_FMT, &fmtdesc)){
            strncpy(fourcc, (char *)&fmtdesc.pixelformat, 4);
            if (fmtdesc.pixelformat == V4L2_PIX_FMT_SGRBG10)
                support_grbg10 = 1;
            c = fmtdesc.flags & 1? 'C' : ' ';
            e = fmtdesc.flags & 2? 'E' : ' ';
            printf("  %s: %c%c %s\n", fourcc, c, e, fmtdesc.description);
            fmtdesc.index++;
    }
    /*
    if (!support_grbg10){
        printf("Doesn't support GRBG10.\n");
        return 1;
    }*/

    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = 640;
    fmt.fmt.pix.height = 480;
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    fmt.fmt.pix.field = V4L2_FIELD_TOP;
    
    if (-1 == xioctl(cap->fd, VIDIOC_S_FMT, &fmt)){
        perror("Setting Pixel Format");
        return 1;
    }

    strncpy(fourcc, (char *)&fmt.fmt.pix.pixelformat, 4);
    printf( "Selected Camera Mode:\n"
            "  Width: %d\n"
            "  Height: %d\n"
            "  PixFmt: %s\n"
            "  Field: %d\n",
            fmt.fmt.pix.width,
            fmt.fmt.pix.height,
            fourcc,
            fmt.fmt.pix.field);
    return 0;
}
 

 
int read_camera(VideoCapture *cap){
    struct v4l2_buffer buf = {0};
    struct timeval tv = {0};
    int r;

    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = 0;
    printf("CAMERA FILE DECSCIPTOR : %d\n", cap->fd);
    if(-1 == xioctl(cap->fd, VIDIOC_QBUF, &buf)){
        perror("Query Buffer");
        return 1;
    }
 
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(cap->fd, &fds);
    
    tv.tv_sec = 2;
    tv.tv_usec = 0;

    r = select(cap->fd + 1, &fds, NULL, NULL, &tv);
    if(-1 == r){
        perror("Waiting for Frame");
        return 1;
    }

    if(-1 == xioctl(cap->fd, VIDIOC_DQBUF, &buf)){
        perror("Retrieving Frame");
        return 1;
    }
    
    return 0;
}

int close_camera(VideoCapture *cap){
    stream_off(cap);
    munmap(cap->data, cap->width * cap->height * (cap->bpp >> 3));
    close(cap->fd);
    return 0;
}